﻿/*
Introduction to the plugin system:

1. Your plugin assemblies go into /build/base/plugins (create folder if needed)
   CAVEAT EMPTOR: Server Updates with the weblauncher might wipe these !!!!

2. Make sure all your interface implementation functions 
   are marked "public" or you get compile errors !!!

3. Server startup and shutdown sequence:
         1. World backup loaded
         2. Simulation started
         3. Plugins created
         4. Init called on all plugins
         5. Start Plugins called on all plugins
         6. Server remains in not-accepting users state until all plugins report they are ready for startup
         7. Game Simulation running, players messing around doing stuff
         8. Shutdown Plugins is called
         9. Server waits for all plug-ins to signal shutdown complete
        10. All plugin instances destroyed. 
        11. world state is shutdown. 
        12. server is shutdown completely
4.  Please keep in mind the difference between returning 
    multiple values in synchronous or asynchronous mode !!!!
    When returning multiple values 
        --> synchronously  using SendRequestToPlugin(...) 
            they get packed into a Lua Array.
        --> asynchronously using SendMessageToPlugin(...) 
            they are returned as multiple arguments in the message event handler !!!!.
        --> Single return values always stay single values.
  
 Intial version of the Tutorial Plugin created by Yorlik July 2018
 Thanks a ton to Zenroth for giving us the plugin system !!!!!
 
 License: MIT License as in 
 https://opensource.org/licenses/MIT
 Citadel NDA has to be respected in any case (as in place July 2018)

 ~Yorlik
 
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
// ---------------------------------------------------------------------------------
// These using are added to the defaults and required for LoA specific Definitions
// ---------------------------------------------------------------------------------
using Gameplay;
using Gameplay.Plugin;
using Gameplay.PublicLua; // This one is only needed if you want to use LuaTableProxy Objects
using System.Collections.Concurrent;
using NLua;


namespace LoaTutorialPlugin
{
    public class TutorialPlugin : IPlugin
    {
        #region internal, private fields
        // This queue is need to send messages back to Lua
        private ConcurrentQueue<LuaResponse> writeQueue;
        // flags if the plugin allows the server to accept users
        private bool isReadyForStartup = false; // blocking server accepting users
        // flags if the plugin is ready for server shutdown
        private bool isReadyForShutdown = false; // blocking server shutdown
        // 
        private DateTime startTime; // set in start function. Just fancy to have ...
        #endregion

        #region IPlugin interface requirements (must implement)
        // You may never change the prototype (arguments and return values) of the functions here
        // It's required by the interface. It might blow up your server if you ignore this.
        // Also you must implement ALL of the functions in this region.
        
        // Called by the server after loading globals, entering the simulation, but NOT yet accepting users
        public void Init(ConcurrentQueue<LuaResponse> _responseQueue)
        {
            Console.WriteLine("Initializing Tutorial Plugin");
            writeQueue = _responseQueue; // Now we have the queue to send messages back to Lua later on
            // Do init stuff - read config files, etc ...
            isReadyForStartup = true; // Allowing users. See IsReadyForServerStartup()
            
        }

        // Called by the server after all plugins have had their Init() method run.
        public void Start()
        {
            // Do Startup stuff. 
            startTime = DateTime.UtcNow;
            Console.WriteLine("Tutorial Plugin started at "+startTime.ToString());
        }

        // prevent shutdown message Spam
        bool signaledShutDown = false;

        // Called by the server startup sequence to determine if users can already be accepted
        public bool IsReadyForServerStartup()
        {
            if ( isReadyForStartup )
            {
                return true;
            }
            else
            {
                return false;
            }
        }

        // executed on server shutdown 
        public void Shutdown()
        {
            if ( !signaledShutDown)
            {
                Console.WriteLine("Shutting down Tutorial Plugin.");
                signaledShutDown = true;
            }
            
            // Doing Shutdown Stuff, like emptying queues, ending tasks, etc ...
            isReadyForShutdown = true; // allowing server to shut down (see IsShutDown() Method)
        }

        // Called by the server shutdown sequence to determine if we can safely shutdown.
        public bool IsShutDown()
        {
            if (isReadyForShutdown)
            {
                //Console.WriteLine("Tutorial Plugin signaling: 'Ready for shutdown'");
                return true;
            }
            else
            {
                return false;
            }
        }

        // Receiving and handling asynchronous requests from Lua here
        // In Lua: First Parameter is your plugin name, second your message Id, after that you just pass whatever you want.
        //         For flexibility you might want to pass a return object ID or a custom event Id for the response.
        //         Data is returned as a Message Event with the eventId given in the LuaResponse.
        //
        // Lua Code:    SendMessageToPlugin("TutorialPlugin", "SomeMessageId", "Data1", 1, true, "More Data", {data = "And even more data"}, ... ) 
        //
        public void MessageFromLua(LuaRequest _request)
        {
            // Async Part 1: Data received from Lua
            // You need to make sure that your Lua Message is shaped properly and has 
            // the correct data types and number of members. Ofc.
            // Don't forget: C# Arrays start with index 0 unlike Lua which starts with 1
            // Normally you'd read data in your according message handler and not right at start 
            // (Except you use a unified message format)
            // - it all depends on your application.
            /* 
             * _request.Data[] has our message payload from Lua
             * e.g.  * /
            ulong  myResponseTargetId = (ulong)(double) _request.Data[0];
            string myResponseEventId  =        (string) _request.Data[1];
             //*/
            

            // Async Part 2: Dispatching Messages by EventId 
            switch(_request.EventId)
            {
                // Async Part 2a: Example one: Return a table
                case "GieffTable":
                {
                    // get data needed for response
                    // Lua has to send it with the Message:
                    // SendMessageToPlugin("TutorialPlugin", "GieffTable", this.Id, "myReturnMessageEventId") 
                    ulong  myResponseTargetId = (ulong)(double) _request.Data[0];
                    string myResponseEventId  =        (string) _request.Data[1];
                    
                    // prepare table
                    LuaTableProxy mTable = new LuaTableProxy();
                    // Add to the HashTable Part
                    mTable.TableData.Add(LuaTableItem.Create("TheAnswer"),   LuaTableItem.Create(42));
                    mTable.TableData.Add(LuaTableItem.Create("TheQuestion"), LuaTableItem.Create("What was it again ... ?"));
                    // Add to the Array Part
                    mTable.TableData.Add(LuaTableItem.Create(1), LuaTableItem.Create("First  Array Element"));
                    mTable.TableData.Add(LuaTableItem.Create(2), LuaTableItem.Create("Second Array Element"));
                    mTable.TableData.Add(LuaTableItem.Create(3), LuaTableItem.Create("Frederick the Great"));
                    
                    // This what we will send back to Lua
                    LuaResponse tableResponse = new LuaResponse();
                    tableResponse.EventId = myResponseEventId; // e.g. myReturnMessageEventId
                    tableResponse.DestinationObjectId = myResponseTargetId; // e.g. this.Id
                    // Response payload
                    tableResponse.Data = new object[1];
                    tableResponse.Data[0] = mTable;
                    writeQueue.Enqueue(tableResponse);  //send message ...
                    break;
                }
                
                // Async Part 2b: Example two: Return a bunch of stuff
                case "GieffStuffs":
                {
                    // get data needed for response
                    // Lua has to send it with the Message:
                    // SendMessageToPlugin("TutorialPlugin", "GieffTable", this.Id, "myReturnMessageEventId") 
                    ulong myResponseTargetId = (ulong)(double)_request.Data[0];
                    string myResponseEventId = (string)_request.Data[1];
                    // prepare table
                    double theAnswer = 42;
                    string advice    = "You must reformulate the question!";
                    string haiku     = "\nNaked\nMonkeys\nRunning\nIn\nCircles.\n";
                    // This what we will send back to Lua
                    LuaResponse mResponse = new LuaResponse();
                    mResponse.EventId             = myResponseEventId;  // e.g. myReturnMessageEventId
                    mResponse.DestinationObjectId = myResponseTargetId; // e.g. this.Id
                    // Response payload
                    mResponse.Data = new object[3]; // make sure you have the correct number of array elements!
                    mResponse.Data[0] = theAnswer;
                    mResponse.Data[1] = advice;
                    mResponse.Data[2] = haiku;
                    writeQueue.Enqueue(mResponse);  //send message ...
                    break;
                }
                
                // Async Part 2c: Catch all - got wrong message 
                default:
                {
                    // BOOM - We received a message we don't understand
                    Console.WriteLine("TestPlugin: Unknown Event Id Reveived: " + _request.EventId.ToString() );
                    break;
                }
                
            }
            
        }

        // Receiving and handling synchronous requests from Lua a.k.a. direct function calls.
        // In Lua: First Parameter is your plugin name, after that you just pass whatever you want.
        //
        // Lua Code:    SendRequestToPlugin("TutorialPlugin", "Data1", 1, true, "More Data", {data = "And even more data"}, ... )  
        // 
        // Data is passed back to Lua as an object array object[] in the return statement, e.g.: return myObject[];
        public object[] DirectRequestFromLua(object[] _requestData)
        {
            // using first piece of data as selector
            string selector = (string)_requestData[0];

            switch (selector)
            {
                #region GimmeFive
                case "GimmeFive":
                    ;
                    var gimme = new object[5];
                    gimme[0] = 1; // These are all elements in a LuaArray once they arrive at the Lua Side
                    gimme[1] = 4;
                    gimme[2] = 9;
                    gimme[3] = 16;
                    gimme[4] = 25;
                    return gimme;
                #endregion
                #region You
                case "You":
                    var too = new object[1];
                    too[0] = "too?"; // This will be a string when it arrives at Lua. 
                                     // No array, since we return only a single element.
                    return too;
                #endregion
                #region Do The Math
                case "DoTheMath":
                    var x = (double)_requestData[1];
                    var y = (double)_requestData[2];
                    var result = new object[4];
                    result[0] = x + y; // These are all elements in a LuaArray once they arrive at the Lua Side
                    result[1] = x - y;
                    result[2] = x * y;
                    result[3] = x / y;
                    return result;
                #endregion
                #region Default error case
                default:
                    object[] returnData = new object[1];
                    returnData[0] = "The answer is 42, but you must reconsider the question";
                    return returnData;
                #endregion
            }
        }

        // Called by the server to determine the name of the plugin.
        public string GetPluginName()
        {
            return "TutorialPlugin";
        }
        #endregion
    }
}
